import RNG from './RNG'

const SCREEN_WIDTH = 64;
const SCREEN_HEIGHT = 32;

const MEMORY_SIZE = 4096;
const STACK_SIZE = 16;

const REGISTERS = 16;

class CPU {
	pc: number; // 16b
	indexReg: number; // 16b
	registers: Uint8Array;

	delayTimer: number; // 8b
	soundTimer: number; // 8b

	stackPointer: number; // 16b
	stack: Array<number>;

	memory: Uint8Array;
	displayMemory: Uint8Array;
	
	rng: RNG;

	constructor() {
		this.memory = new Uint8Array(MEMORY_SIZE);
		this.displayMemory = new Uint8Array(SCREEN_WIDTH*SCREEN_HEIGHT);
		this.pc = 0x200;
		this.indexReg = 0;
		this.stackPointer = 0;
		this.stack = new Array<number>(STACK_SIZE);
		this.delayTimer = 0;
		this.soundTimer = 0;
		this.registers = new Uint8Array(REGISTERS);
		
		this.rng = new RNG(0);

		this.clearMemory();
		this.clearDisplay();
		this.clearRegisters();

		// 0x000-0x1FF - Chip 8 interpreter
		// 0x050-0x0A0 - 4x5 pixel font set (0-F)
		// 0x200-0xFFF - Program ROM and work RAM
		this.setupFont();
	}

	clearRegisters() {
		this.registers.fill(0);
		this.pc = 0x200; // Start of the program on normal programs
		this.indexReg = 0;
		this.stackPointer = 0;
		this.delayTimer = 0;
		this.soundTimer = 0;

		this.rng = new RNG(0);
	}

	clearMemory() {
		this.memory.fill(0);
	}

	clearDisplay() {
		this.displayMemory.fill(0);
	}


	private setupFont() {
		// Using address from 050 - 09F ??
		let addressFont = 0x050;
		
		// Every character has 5 Bytes
		const characters = [
			0xF0, 0x90, 0x90, 0x90, 0xF0, // 0
			0x20, 0x60, 0x20, 0x20, 0x70, // 1
			0xF0, 0x10, 0xF0, 0x80, 0xF0, // 2
			0xF0, 0x10, 0xF0, 0x10, 0xF0, // 3
			0x90, 0x90, 0xF0, 0x10, 0x10, // 4
			0xF0, 0x80, 0xF0, 0x10, 0xF0, // 5
			0xF0, 0x80, 0xF0, 0x90, 0xF0, // 6
			0xF0, 0x10, 0x20, 0x40, 0x40, // 7
			0xF0, 0x90, 0xF0, 0x90, 0xF0, // 8
			0xF0, 0x90, 0xF0, 0x10, 0xF0, // 9
			0xF0, 0x90, 0xF0, 0x90, 0x90, // A
			0xE0, 0x90, 0xE0, 0x90, 0xE0, // B
			0xF0, 0x80, 0x80, 0x80, 0xF0, // C
			0xE0, 0x90, 0x90, 0x90, 0xE0, // D
			0xF0, 0x80, 0xF0, 0x80, 0xF0, // E
			0xF0, 0x80, 0xF0, 0x80, 0x80  // F
		];

		this.memory.set(characters, addressFont);
	}

	private incrementPC() {
		this.pc += 2;
	}
	
	private cycle() {
		// Fetch opcode
		let firstOpcode: number = this.memory[this.pc];
		let secondtOpcode: number = this.memory[this.pc + 1];
		let opcode: number = firstOpcode << 8 | secondtOpcode;
		
		// Decode opcode

		let firstNibble = opcode >> 12;

		switch(firstNibble) {

		}
	}

	
}

export default CPU;